@@ -36,16 +36,10 @@
 #include "key.h"
 #include "string.h"
 #include "random.h"
-#include "key-private.h"
+#include "missing.h"
 
 #ifndef KEYCTL_DH_COMPUTE
 #define KEYCTL_DH_COMPUTE 23
-
-struct keyctl_dh_params {
-	int32_t private;
-	int32_t prime;
-	int32_t base;
-};
 #endif
 
 #ifndef KEYCTL_PKEY_QUERY
@@ -80,6 +74,20 @@ struct keyctl_pkey_params {
 	};
 	uint32_t __spare[7];
 };
+
+/* Work around the missing (pre-4.7) or broken (4.14.{70,71,72} and
+ * 4.18.{8,9,10}) kernel declaration of struct keyctl_dh_params
+ */
+struct dh_params {
+	int32_t private;
+	int32_t prime;
+	int32_t base;
+};
+#else
+/* When KEYCTL_PKEY_QUERY is defined by the kernel, the
+ * struct keyctl_dh_params declaration is valid.
+ */
+#define dh_params keyctl_dh_params
 #endif
 
 #ifndef KEYCTL_RESTRICT_KEYRING
@@ -131,11 +139,11 @@ static long kernel_update_key(int32_t serial, const void *payload, size_t len)
 	return result >= 0 ? result : -errno;
 }
 
-static long kernel_revoke_key(int32_t serial)
+static long kernel_invalidate_key(int32_t serial)
 {
 	long result;
 
-	result = syscall(__NR_keyctl, KEYCTL_REVOKE, serial);
+	result = syscall(__NR_keyctl, KEYCTL_INVALIDATE, serial);
 
 	return result >= 0 ? result : -errno;
 }
@@ -202,9 +210,9 @@ static long kernel_dh_compute(int32_t private, int32_t prime, int32_t base,
 {
 	long result;
 
-	struct keyctl_dh_params params = { .private = private,
-					   .prime = prime,
-					   .base = base };
+	struct dh_params params = { .private = private,
+				    .prime = prime,
+				    .base = base };
 
 	result = syscall(__NR_keyctl, KEYCTL_DH_COMPUTE, &params, payload, len,
 			NULL);
@@ -241,6 +249,26 @@ static long kernel_key_eds(int op, int32_t serial, const char *encoding,
 	return result >= 0 ? result : -errno;
 }
 
+static long kernel_key_verify(int32_t serial,
+				const char *encoding, const char *hash,
+				const void *data, size_t data_len,
+				const void *sig, size_t sig_len)
+{
+	struct keyctl_pkey_params params = {
+		.key_id		= serial,
+		.in_len		= data_len,
+		.in2_len	= sig_len,
+	};
+	char *info = format_key_info(encoding, hash);
+	long result;
+
+	result = syscall(__NR_keyctl, KEYCTL_PKEY_VERIFY, &params,
+				info ?: "", data, sig);
+	l_free(info);
+
+	return result >= 0 ? result : -errno;
+}
+
 static bool setup_internal_keyring(void)
 {
 	internal_keyring = kernel_add_key("keyring", "ell-internal", NULL, 0,
@@ -259,6 +287,7 @@ LIB_EXPORT struct l_key *l_key_new(enum l_key_type type, const void *payload,
 {
 	struct l_key *key;
 	char *description;
+	static unsigned long key_idx;
 
 	if (unlikely(!payload))
 		return NULL;
@@ -271,7 +300,7 @@ LIB_EXPORT struct l_key *l_key_new(enum l_key_type type, const void *payload,
 
 	key = l_new(struct l_key, 1);
 	key->type = type;
-	description = l_strdup_printf("ell-key-%p", key);
+	description = l_strdup_printf("ell-key-%lu", key_idx++);
 	key->serial = kernel_add_key(key_type_names[type], description, payload,
 					payload_length, internal_keyring);
 	l_free(description);
@@ -296,7 +325,12 @@ LIB_EXPORT void l_key_free(struct l_key *key)
 	if (unlikely(!key))
 		return;
 
-	kernel_revoke_key(key->serial);
+	/*
+	 * Use invalidate as, unlike revoke, this doesn't delay the
+	 * key garbage collection and causes the quota used by the
+	 * key to be released sooner and more predictably.
+	 */
+	kernel_invalidate_key(key->serial);
 
 	l_free(key);
 }
@@ -333,7 +367,7 @@ LIB_EXPORT bool l_key_extract(struct l_key *key, void *payload, size_t *len)
 	keylen = kernel_read_key(key->serial, payload, *len);
 
 	if (keylen < 0 || (size_t)keylen > *len) {
-		memset(payload, 0, *len);
+		explicit_bzero(payload, *len);
 		return false;
 	}
 
@@ -352,10 +386,7 @@ static const char *lookup_cipher(enum l_key_cipher_type cipher)
 
 	switch (cipher) {
 	case L_KEY_RSA_PKCS1_V1_5:
-		/* Padding is handled in userspace, so the kernel only sees
-		 * raw RSA operations
-		 */
-		ret = "raw";
+		ret = "pkcs1";
 		break;
 	case L_KEY_RSA_RAW:
 		ret = "raw";
@@ -405,25 +436,66 @@ LIB_EXPORT bool l_key_get_info(struct l_key *key, enum l_key_cipher_type cipher,
 	if (unlikely(!key))
 		return false;
 
-	/* Other checksum types are not yet supported */
-	if (checksum != L_CHECKSUM_NONE)
-		return false;
-
 	return !kernel_query_key(key->serial, lookup_cipher(cipher),
 					lookup_checksum(checksum), bits,
 					public);
 }
 
-static bool compute_common(struct l_key *base,
-			   struct l_key *private,
-			   struct l_key *prime,
-			   void *payload, size_t *len)
+LIB_EXPORT struct l_key *l_key_generate_dh_private(const void *prime_buf,
+							size_t prime_len)
+{
+	uint8_t *buf;
+	const uint8_t *prime = prime_buf;
+	size_t prime_bits;
+	unsigned int i;
+	size_t private_bytes;
+	size_t random_bytes;
+	struct l_key *private;
+
+	/* Find the prime's bit length excluding leading 0s */
+
+	for (i = 0; i < prime_len && !prime[i]; i++);
+
+	if (i == prime_len || (i == prime_len - 1 && prime[i] < 5))
+		return NULL;
+
+	prime_bits = (prime_len - i) * 8 - __builtin_clz(prime[i]);
+
+	/*
+	 * Generate a random DH private value conforming to 1 < x < p - 1.
+	 * To do this covering all possible values in this range with the
+	 * same probability of generating each value generally requires
+	 * looping.  Instead we generate a value in the range
+	 * [2 ^ (prime_bits - 2), 2 ^ (prime_bits - 1) - 1] by forcing bit
+	 * prime_bits - 2 to 1, i.e. the range in PKCS #3 Section 7.1 for
+	 * l equal to prime_bits - 1.  This means we're using between
+	 * one half and one quarter of the full [2, p - 2] range, i.e.
+	 * between 1 and 2 bits fewer.  Note that since p is odd
+	 * p - 1 has the same bit length as p and so our maximum value
+	 * 2 ^ (prime_bits - 1) - 1 is still less than p - 1.
+	 */
+	private_bytes = ((prime_bits - 1) + 7) / 8;
+	random_bytes = ((prime_bits - 2) + 7) / 8;
+	buf = l_malloc(private_bytes);
+	l_getrandom(buf + private_bytes - random_bytes, random_bytes);
+
+	buf[0] &= (1 << ((prime_bits - 2) % 8)) - 1;
+	buf[0] |= 1 << ((prime_bits - 2) % 8);
+
+	private = l_key_new(L_KEY_RAW, buf, private_bytes);
+	explicit_bzero(buf, private_bytes);
+	l_free(buf);
+	return private;
+}
+
+static bool compute_common(struct l_key *base, struct l_key *private,
+				struct l_key *prime, void *payload, size_t *len)
 {
 	long result_len;
 	bool usable_payload = *len != 0;
 
 	result_len = kernel_dh_compute(private->serial, prime->serial,
-				       base->serial, payload, *len);
+					base->serial, payload, *len);
 
 	if (result_len > 0) {
 		*len = result_len;
@@ -449,6 +521,59 @@ LIB_EXPORT bool l_key_compute_dh_secret(struct l_key *other_public,
 	return compute_common(other_public, private, prime, payload, len);
 }
 
+static int be_bignum_compare(const uint8_t *a, size_t a_len,
+				const uint8_t *b, size_t b_len)
+{
+	unsigned int i;
+
+	if (a_len >= b_len) {
+		for (i = 0; i < a_len - b_len; i++)
+			if (a[i])
+				return 1;
+
+		return memcmp(a + i, b, b_len);
+	} else {
+		for (i = 0; i < b_len - a_len; i++)
+			if (b[i])
+				return -1;
+
+		return memcmp(a, b + i, a_len);
+	}
+}
+
+/*
+ * Validate that @payload is within range for a private and public key for
+ * a DH computation in the finite field group defined by modulus @prime_buf,
+ * both numbers stored as big-endian integers.  We require a key in the
+ * [2, prime - 2] (inclusive) interval.  PKCS #3 does not exclude 1 as a
+ * private key but other specs do.
+ */
+LIB_EXPORT bool l_key_validate_dh_payload(const void *payload, size_t len,
+				const void *prime_buf, size_t prime_len)
+{
+	static const uint8_t one[] = { 1 };
+	uint8_t prime_1[prime_len];
+
+	/*
+	 * Produce prime - 1 for the payload < prime - 1 check.
+	 * prime is odd so just zero the LSB.
+	 */
+	memcpy(prime_1, prime_buf, prime_len);
+
+	if (prime_len < 1 || !(prime_1[prime_len - 1] & 1))
+		return false;
+
+	prime_1[prime_len - 1] &= ~1;
+
+	if (be_bignum_compare(payload, len, one, 1) <= 0)
+		return false;
+
+	if (be_bignum_compare(payload, len, prime_1, prime_len) >= 0)
+		return false;
+
+	return true;
+}
+
 /* Common code for encrypt/decrypt/sign */
 static ssize_t eds_common(struct l_key *key,
 				enum l_key_cipher_type cipher,
@@ -464,102 +589,18 @@ static ssize_t eds_common(struct l_key *key,
 				len_out);
 }
 
-static void getrandom_nonzero(uint8_t *buf, int len)
-{
-	l_getrandom(buf, len);
-
-	while (len--) {
-		while (buf[0] == 0)
-			l_getrandom(buf, 1);
-
-		buf++;
-	}
-}
-
-/* PKCS#1 v1.5 RSA padding according to RFC3447 */
-static uint8_t *pad(const uint8_t *in, size_t len_in, size_t len_out,
-			uint8_t flag, bool randfill)
-{
-	size_t fill_len;
-	uint8_t *padded;
-
-	if (len_in > len_out - 11)
-		return NULL;
-
-	padded = l_malloc(len_out);
-	fill_len = len_out - len_in - 3;
-
-	padded[0] = 0x00;
-	padded[1] = flag;
-
-	if (randfill)
-		getrandom_nonzero(padded + 2, fill_len);
-	else
-		memset(padded + 2, 0xff, fill_len);
-
-	padded[fill_len + 2] = 0x00;
-	memcpy(padded + fill_len + 3, in, len_in);
-
-	return padded;
-}
-
-static ssize_t unpad(uint8_t *in, uint8_t *out, size_t len_in, size_t len_out,
-			uint8_t flag, bool randfill)
-{
-	size_t pos;
-	ssize_t unpad_len;
-
-	if (in[0] != 0x00 || in[1] != flag)
-		return -EINVAL;
-
-	if (randfill)
-		for (pos = 2; pos < len_in && in[pos] != 0x00; pos++);
-	else
-		for (pos = 2; pos < len_in && in[pos] == 0xff; pos++);
-
-	if (pos < 10 || pos == len_in)
-		return -EINVAL;
-
-	pos++;
-	unpad_len = len_in - pos;
-
-	if (out) {
-		if ((size_t)unpad_len > len_out)
-			return -EINVAL;
-
-		memcpy(out, in + pos, unpad_len);
-	}
-
-	return unpad_len;
-}
-
 LIB_EXPORT ssize_t l_key_encrypt(struct l_key *key,
 					enum l_key_cipher_type cipher,
 					enum l_checksum_type checksum,
 					const void *in, void *out,
 					size_t len_in, size_t len_out)
 {
-	uint8_t *padded = NULL;
 	ssize_t ret_len;
 
-	/* Other checksum types are not yet supported */
-	if (checksum != L_CHECKSUM_NONE)
-		return -EINVAL;
-
-	if (cipher == L_KEY_RSA_PKCS1_V1_5) {
-		padded = pad(in, len_in, len_out, 0x02, true);
-		if (!padded)
-			return -EOVERFLOW;
-
-		cipher = L_KEY_RSA_RAW;
-	}
-
-	ret_len = eds_common(key, cipher, checksum, padded ?: in, out,
-				padded ? len_out : len_in, len_out,
+	ret_len = eds_common(key, cipher, checksum, in, out,
+				len_in, len_out,
 				KEYCTL_PKEY_ENCRYPT);
 
-	l_free(padded);
-
 	return ret_len;
 }
 
@@ -569,28 +610,15 @@ LIB_EXPORT ssize_t l_key_decrypt(struct l_key *key,
 					const void *in, void *out,
 					size_t len_in, size_t len_out)
 {
-	uint8_t *padded = NULL;
 	ssize_t ret_len;
 
-	/* Other checksum types are not yet supported */
-	if (checksum != L_CHECKSUM_NONE)
-		return -EINVAL;
-
-	if (cipher == L_KEY_RSA_PKCS1_V1_5)
-		padded = l_malloc(len_in);
-
-	ret_len = eds_common(key, cipher, checksum, in, padded ?: out, len_in,
-				padded ? len_in : len_out, KEYCTL_PKEY_DECRYPT);
+	ret_len = eds_common(key, cipher, checksum, in, out, len_in,
+				len_out, KEYCTL_PKEY_DECRYPT);
 
 	if (ret_len < 0)
 		goto done;
 
-	if (padded)
-		ret_len = unpad(padded, out, ret_len, len_out, 0x02, true);
-
 done:
-	l_free(padded);
-
 	return ret_len;
 }
 
@@ -599,27 +627,12 @@ LIB_EXPORT ssize_t l_key_sign(struct l_key *key,
 				enum l_checksum_type checksum, const void *in,
 				void *out, size_t len_in, size_t len_out)
 {
-	uint8_t *padded = NULL;
 	ssize_t ret_len;
 
-	/* Other checksum types are not yet supported */
-	if (checksum != L_CHECKSUM_NONE)
-		return -EINVAL;
-
-	if (cipher == L_KEY_RSA_PKCS1_V1_5) {
-		padded = pad(in, len_in, len_out, 0x01, false);
-		if (!padded)
-			return -EOVERFLOW;
-
-		cipher = L_KEY_RSA_RAW;
-	}
-
-	ret_len = eds_common(key, cipher, checksum, padded ?: in, out,
-				padded ? len_out : len_in, len_out,
+	ret_len = eds_common(key, cipher, checksum, in, out,
+				len_in, len_out,
 				KEYCTL_PKEY_SIGN);
 
-	l_free(padded);
-
 	return ret_len;
 }
 
@@ -629,60 +642,30 @@ LIB_EXPORT bool l_key_verify(struct l_key *key,
 				const void *sig, size_t len_data,
 				size_t len_sig)
 {
-	enum l_key_cipher_type kernel_cipher;
-	ssize_t hash_len;
-	uint8_t *compare_hash;
-	L_AUTO_FREE_VAR(uint8_t *, sig_hash);
+	long result;
 
-	sig_hash = l_malloc(len_sig);
-
-	/* Other checksum types are not yet supported */
-	if (checksum != L_CHECKSUM_NONE)
+	if (unlikely(!key))
 		return false;
 
-	/* The keyctl verify implementation compares the verify results
-	 * before we get a chance to unpad it. Instead, use the *encrypt*
-	 * operation (which uses the same math as verify) to get the hash
-	 * returned to us so it can be unpadded before comparing
-	 */
-	if (cipher == L_KEY_RSA_PKCS1_V1_5)
-		kernel_cipher = L_KEY_RSA_RAW;
-	else
-		kernel_cipher = cipher;
+	result = kernel_key_verify(key->serial, lookup_cipher(cipher),
+					lookup_checksum(checksum),
+					data, len_data,
+					sig, len_sig);
 
-	hash_len = eds_common(key, kernel_cipher, checksum, sig, sig_hash,
-				len_sig, len_sig, KEYCTL_PKEY_ENCRYPT);
-
-	if (hash_len < 0)
-		return false;
-
-	compare_hash = sig_hash;
-
-	if (cipher == L_KEY_RSA_PKCS1_V1_5) {
-		ssize_t unpad_len;
-
-		unpad_len = unpad(sig_hash, NULL, hash_len, 0, 0x01, false);
-		if (unpad_len < 0)
-			return false;
-
-		compare_hash += hash_len - unpad_len;
-		hash_len = unpad_len;
-	}
-
-	return (len_data == (size_t)hash_len) &&
-		(memcmp(data, compare_hash, hash_len) == 0);
+	return result >= 0;
 }
 
 LIB_EXPORT struct l_keyring *l_keyring_new(void)
 {
 	struct l_keyring *keyring;
 	char *description;
+	static unsigned long keyring_idx;
 
 	if (!internal_keyring && !setup_internal_keyring())
 		return NULL;
 
 	keyring = l_new(struct l_keyring, 1);
-	description = l_strdup_printf("ell-keyring-%p", keyring);
+	description = l_strdup_printf("ell-keyring-%lu", keyring_idx++);
 	keyring->serial = kernel_add_key("keyring", description, NULL, 0,
 						internal_keyring);
 	l_free(description);
@@ -735,7 +718,7 @@ LIB_EXPORT void l_keyring_free(struct l_keyring *keyring)
 	if (unlikely(!keyring))
 		return;
 
-	kernel_revoke_key(keyring->serial);
+	kernel_invalidate_key(keyring->serial);
 
 	l_free(keyring);
 }
@@ -750,7 +733,8 @@ LIB_EXPORT void l_keyring_free_norevoke(struct l_keyring *keyring)
 	l_free(keyring);
 }
 
-bool l_keyring_link(struct l_keyring *keyring, const struct l_key *key)
+LIB_EXPORT bool l_keyring_link(struct l_keyring *keyring,
+							const struct l_key *key)
 {
 	long error;
 
@@ -762,7 +746,8 @@ bool l_keyring_link(struct l_keyring *keyring, const struct l_key *key)
 	return error == 0;
 }
 
-bool l_keyring_unlink(struct l_keyring *keyring, const struct l_key *key)
+LIB_EXPORT bool l_keyring_unlink(struct l_keyring *keyring,
+							const struct l_key *key)
 {
 	long error;
 
@@ -774,7 +759,33 @@ bool l_keyring_unlink(struct l_keyring *keyring, const struct l_key *key)
 	return error == 0;
 }
 
-bool l_key_is_supported(uint32_t features)
+LIB_EXPORT bool l_keyring_link_nested(struct l_keyring *keyring,
+						const struct l_keyring *nested)
+{
+	long error;
+
+	if (unlikely(!keyring) || unlikely(!nested))
+		return false;
+
+	error = kernel_link_key(nested->serial, keyring->serial);
+
+	return error == 0;
+}
+
+LIB_EXPORT bool l_keyring_unlink_nested(struct l_keyring *keyring,
+						const struct l_keyring *nested)
+{
+	long error;
+
+	if (unlikely(!keyring) || unlikely(!nested))
+		return false;
+
+	error = kernel_unlink_key(nested->serial, keyring->serial);
+
+	return error == 0;
+}
+
+LIB_EXPORT bool l_key_is_supported(uint32_t features)
 {
 	long result;
 
