@@ -35,6 +35,11 @@
 extern "C" {
 #endif
 
+#define l_container_of(ptr, type, member) ({				\
+		const __typeof__(((type *) 0)->member) *__mptr = (ptr);	\
+		(type *)((char *) __mptr - offsetof(type, member));	\
+	})
+
 #define likely(x)   __builtin_expect(!!(x), 1)
 #define unlikely(x) __builtin_expect(!!(x), 0)
 
@@ -102,6 +107,10 @@ do {						\
 #error "Unknown byte order"
 #endif
 
+#if __STDC_VERSION__ <= 199409L
+#define inline __inline__
+#endif
+
 static inline uint8_t l_get_u8(const void *ptr)
 {
 	return *((const uint8_t *) ptr);
@@ -239,6 +248,14 @@ static inline void auto_free(void *a)
 	l_free(*p);
 }
 
+static inline size_t minsize(size_t a, size_t b)
+{
+	if (a <= b)
+		return a;
+
+	return b;
+}
+
 /**
  * l_new:
  * @type: type of structure
@@ -268,6 +285,7 @@ bool l_str_has_prefix(const char *str, const char *prefix);
 bool l_str_has_suffix(const char *str, const char *suffix);
 
 char *l_util_hexstring(const unsigned char *buf, size_t len);
+char *l_util_hexstring_upper(const unsigned char *buf, size_t len);
 unsigned char *l_util_from_hexstring(const char *str, size_t *out_len);
 
 typedef void (*l_util_hexdump_func_t) (const char *str, void *user_data);
