@@ -24,12 +24,15 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
+
 #include <ell/ell.h>
 
-#include "eap.h"
-#include "eap-private.h"
-#include "util.h"
-#include "ecc.h"
+#include "src/missing.h"
+#include "src/eap.h"
+#include "src/eap-private.h"
+#include "src/crypto.h"
+#include "src/util.h"
 
 #define EAP_PWD_GROUP_DESC	19
 #define EAP_PWD_RAND_FN		0x01
@@ -65,13 +68,14 @@ struct eap_pwd_handle {
 	enum eap_pwd_prep prep;
 	char *identity;
 	char *password;
-	struct ecc_point pwe;
-	struct ecc_point element_s;
-	struct ecc_point element_p;
+	const struct l_ecc_curve *curve;
+	struct l_ecc_point *pwe;
+	struct l_ecc_point *element_s;
+	struct l_ecc_point *element_p;
 	uint32_t ciphersuite;
-	uint64_t scalar_s[NUM_ECC_DIGITS];
-	uint64_t scalar_p[NUM_ECC_DIGITS];
-	uint64_t p_rand[NUM_ECC_DIGITS];
+	struct l_ecc_scalar *scalar_s;
+	struct l_ecc_scalar *scalar_p;
+	struct l_ecc_scalar *p_rand;
 	uint8_t *rx_frag_buf;
 	uint16_t rx_frag_total;
 	uint16_t rx_frag_count;
@@ -80,37 +84,6 @@ struct eap_pwd_handle {
 	uint16_t tx_frag_remaining;
 };
 
-static uint64_t curve_p[NUM_ECC_DIGITS] = CURVE_P_32;
-
-static bool H(uint8_t num_args, uint8_t *out, ...)
-{
-	struct l_checksum *hmac;
-	struct iovec iov[num_args];
-	uint8_t k[32] = { 0 };
-	va_list va;
-	int i;
-	int ret;
-
-	va_start(va, out);
-
-	hmac = l_checksum_new_hmac(L_CHECKSUM_SHA256, k, 32);
-	if (!hmac)
-		return false;
-
-	for (i = 0; i < num_args; i++) {
-		iov[i].iov_base = va_arg(va, void *);
-		iov[i].iov_len = va_arg(va, size_t);
-	}
-
-	if (!l_checksum_updatev(hmac, iov, num_args))
-		return false;
-
-	ret = l_checksum_get_digest(hmac, out, 32);
-	l_checksum_free(hmac);
-
-	return (ret == 32);
-}
-
 /* RFC 5931, Section 2.5 - Key Derivation Function */
 static bool kdf(uint8_t *key, size_t key_len, const char *label,
 		size_t label_len, void *out, size_t olen)
@@ -145,7 +118,7 @@ static bool kdf(uint8_t *key, size_t key_len, const char *label,
 		if (!l_checksum_updatev(hmac, iov, iov_pos))
 			return false;
 
-		l_checksum_get_digest(hmac, out + len, 32);
+		l_checksum_get_digest(hmac, out + len, minsize(olen - len, 32));
 		l_checksum_free(hmac);
 
 		len += 32;
@@ -172,13 +145,20 @@ static bool eap_pwd_reset_state(struct eap_state *eap)
 	pwd->rx_frag_total = 0;
 
 	pwd->prep = EAP_PWD_PREP_NONE;
-	memset(&pwd->pwe, 0, sizeof(struct ecc_point));
-	memset(&pwd->element_s, 0, sizeof(struct ecc_point));
-	memset(&pwd->element_p, 0, sizeof(struct ecc_point));
 	pwd->ciphersuite = 0;
-	memset(pwd->scalar_s, 0, sizeof(pwd->scalar_s));
-	memset(pwd->scalar_p, 0, sizeof(pwd->scalar_p));
-	memset(pwd->p_rand, 0, sizeof(pwd->p_rand));
+
+	l_ecc_point_free(pwd->pwe);
+	pwd->pwe = NULL;
+	l_ecc_point_free(pwd->element_p);
+	pwd->element_p = NULL;
+	l_ecc_point_free(pwd->element_s);
+	pwd->element_s = NULL;
+	l_ecc_scalar_free(pwd->scalar_p);
+	pwd->scalar_p = NULL;
+	l_ecc_scalar_free(pwd->scalar_s);
+	pwd->scalar_s = NULL;
+	l_ecc_scalar_free(pwd->p_rand);
+	pwd->p_rand = NULL;
 
 	return true;
 }
@@ -189,7 +169,12 @@ static void eap_pwd_free(struct eap_state *eap)
 
 	eap_pwd_reset_state(eap);
 	l_free(pwd->identity);
-	l_free(pwd->password);
+
+	if (pwd->password) {
+		explicit_bzero(pwd->password, strlen(pwd->password));
+		l_free(pwd->password);
+	}
+
 	l_free(pwd);
 
 	eap_set_data(eap, NULL);
@@ -250,12 +235,13 @@ static void eap_pwd_handle_id(struct eap_state *eap,
 	uint8_t rand_fn;
 	uint8_t prf;
 	uint32_t token;
-	uint8_t counter = 1;
+	uint8_t counter = 0;
 	uint8_t resp[15 + strlen(pwd->identity)];
 	uint8_t *pos;
-	uint8_t pwd_seed[ECC_BYTES];
-	uint64_t pwd_value[NUM_ECC_DIGITS];	/* used as X value */
-	uint64_t y_value[NUM_ECC_DIGITS];
+	uint8_t pwd_seed[32];
+	uint8_t pwd_value[L_ECC_SCALAR_MAX_BYTES];	/* used as X value */
+	size_t nbytes;
+	bool found = false;
 
 	/*
 	 * Group desc (2) + Random func (1) + prf (1) + token (4) + prep (1) +
@@ -274,7 +260,9 @@ static void eap_pwd_handle_id(struct eap_state *eap,
 	pwd->state = EAP_PWD_STATE_ID;
 
 	group = l_get_be16(pkt);
-	if (group != EAP_PWD_GROUP_DESC) {
+
+	pwd->curve = l_ecc_curve_get_ike_group(group);
+	if (!pwd->curve) {
 		l_error("group %d not supported", group);
 		goto error;
 	}
@@ -308,11 +296,18 @@ static void eap_pwd_handle_id(struct eap_state *eap,
 		goto error;
 	}
 
+	nbytes = l_ecc_curve_get_scalar_bytes(pwd->curve);
+
 	while (counter < 20) {
+		struct l_ecc_point *pwe = NULL;
+
+		counter++;
+
 		/* pwd-seed = H(token|peer-ID|server-ID|password|counter) */
-		H(5, pwd_seed, &token, 4, pwd->identity, strlen(pwd->identity),
-				pkt + 9, len - 9, pwd->password,
-				strlen(pwd->password), &counter, 1);
+		hkdf_extract(L_CHECKSUM_SHA256, NULL, 0, 5, pwd_seed, &token, 4,
+				pwd->identity, strlen(pwd->identity), pkt + 9,
+				len - 9, pwd->password, strlen(pwd->password),
+				&counter, 1);
 
 		/*
 		 * pwd-value = KDF(pwd-seed, "EAP-pwd Hunting And Pecking",
@@ -320,29 +315,30 @@ static void eap_pwd_handle_id(struct eap_state *eap,
 		 */
 		kdf(pwd_seed, 32, "EAP-pwd Hunting And Pecking",
 				strlen("EAP-pwd Hunting And Pecking"),
-				pwd_value, 32);
+				pwd_value, nbytes);
 
-		ecc_be2native(pwd_value);
+		if (!(pwd_seed[31] & 1))
+			pwe = l_ecc_point_from_data(pwd->curve,
+					L_ECC_POINT_TYPE_COMPRESSED_BIT1,
+					pwd_value, nbytes);
+		else
+			pwe = l_ecc_point_from_data(pwd->curve,
+					L_ECC_POINT_TYPE_COMPRESSED_BIT0,
+					pwd_value, nbytes);
 
-		if (ecc_compute_y(y_value, pwd_value)) {
-			l_info("computed y in %u tries", counter);
+		if (!pwe)
+			continue;
 
-			/* unambiguously choose Y coordinate */
-			if ((y_value[0] & 1) != (pwd_seed[31] & 1))
-				vli_mod_sub(y_value, curve_p, y_value, curve_p);
-
-			memcpy(pwd->pwe.x, pwd_value, 32);
-			memcpy(pwd->pwe.y, y_value, 32);
-
-			if (!ecc_valid_point(&pwd->pwe))
-				goto invalid_point;
-
-			break;
-		}
-
-		counter++;
+		if (!found) {
+			found = true;
+			pwd->pwe = pwe;
+		} else
+			l_ecc_point_free(pwe);
 	}
 
+	explicit_bzero(pwd_seed, sizeof(pwd_seed));
+	explicit_bzero(pwd_value, sizeof(pwd_value));
+
 	pos = resp + 5; /* header */
 	*pos++ = EAP_PWD_EXCH_ID;
 	l_put_be16(group, pos);
@@ -359,8 +355,6 @@ static void eap_pwd_handle_id(struct eap_state *eap,
 
 	return;
 
-invalid_point:
-	l_error("point not on curve");
 error:
 	eap_method_error(eap);
 }
@@ -369,14 +363,16 @@ static void eap_pwd_handle_commit(struct eap_state *eap,
 					const uint8_t *pkt, size_t len)
 {
 	struct eap_pwd_handle *pwd = eap_get_data(eap);
-	uint8_t resp[102];
+	uint8_t resp[L_ECC_POINT_MAX_BYTES + L_ECC_SCALAR_MAX_BYTES + 6];
 	uint8_t *pos;
-	uint64_t p_mask[NUM_ECC_DIGITS];
-	uint64_t one[NUM_ECC_DIGITS] = { 1 };
-	uint64_t curve_n[NUM_ECC_DIGITS] = CURVE_N_32;
+	struct l_ecc_scalar *p_mask;
+	struct l_ecc_scalar *order;
+	size_t nbytes = l_ecc_curve_get_scalar_bytes(pwd->curve);
 
-	if (len != 96) {
-		l_error("bad packet length, expected 96, got %zu", len);
+	/* [Element (nbytes * 2)][Scalar (nbytes)] */
+	if (len != nbytes + nbytes * 2) {
+		l_error("bad packet length, expected %zu, got %zu",
+				nbytes + nbytes * 2, len);
 		goto error;
 	}
 
@@ -388,73 +384,40 @@ static void eap_pwd_handle_commit(struct eap_state *eap,
 	pwd->state = EAP_PWD_STATE_COMMIT;
 
 	/*
-	 * RFC 5114 Section 2.6 - 256-bit Random ECP Group
-	 * Prime p is 32 bytes in length, therefore x and y will also each be
-	 * 32 bytes in length (total of 64), leaving the remainder for the
-	 * scalar value (32).
+	 * Commit contains Element_S (nbytes * 2) then Scalar_s (nbytes)
 	 */
-	memcpy(pwd->element_s.x, pkt, ECC_BYTES);
-	memcpy(pwd->element_s.y, pkt + ECC_BYTES, ECC_BYTES);
-	memcpy(pwd->scalar_s, pkt + 64, ECC_BYTES);
-
-	ecc_be2native(pwd->element_s.x);
-	ecc_be2native(pwd->element_s.y);
-	ecc_be2native(pwd->scalar_s);
-
-	if (!ecc_valid_point(&pwd->element_s))
+	pwd->element_s = l_ecc_point_from_data(pwd->curve,
+						L_ECC_POINT_TYPE_FULL,
+						pkt, nbytes * 2);
+	if (!pwd->element_s)
 		goto invalid_point;
 
-	/*
-	 * RFC 5931 Section 2.8.4.1
-	 *
-	 * chose two random numbers, 1 < s_rand, s_mask < r
-	 * compute Scalar_P and Element_P
-	 * Scalar_P = (p_rand + p_mask) mod r
-	 * Element_P = inv(p_mask * PWE)
-	 */
-	l_getrandom(pwd->p_rand, ECC_BYTES);
+	pwd->scalar_s = l_ecc_scalar_new(pwd->curve, pkt + nbytes * 2, nbytes);
 
-	/* ensure 1 < p_rand < r */
-	while (!((vli_cmp(pwd->p_rand, one) > 0) &&
-			(vli_cmp(pwd->p_rand, curve_n) < 0)))
-		l_getrandom(pwd->p_rand, ECC_BYTES);
+	pwd->p_rand = l_ecc_scalar_new_random(pwd->curve);
+	p_mask = l_ecc_scalar_new_random(pwd->curve);
+	pwd->scalar_p = l_ecc_scalar_new(pwd->curve, NULL, 0);
 
-	l_getrandom(p_mask, ECC_BYTES);
+	order = l_ecc_curve_get_order(pwd->curve);
 
-	/* ensure 1 < p_mask < r */
-	while (!((vli_cmp(p_mask, one) > 0) &&
-			(vli_cmp(p_mask, curve_n) < 0)))
-		l_getrandom(p_mask, ECC_BYTES);
+	l_ecc_scalar_add(pwd->scalar_p, pwd->p_rand, p_mask, order);
 
-	vli_mod_add(pwd->scalar_p, pwd->p_rand, p_mask, curve_n);
+	l_ecc_scalar_free(order);
 
+	pwd->element_p = l_ecc_point_new(pwd->curve);
 	/* p_mask * PWE */
-	ecc_point_mult(&pwd->element_p, &pwd->pwe, p_mask, NULL,
-		vli_num_bits(p_mask));
+	l_ecc_point_multiply(pwd->element_p, p_mask, pwd->pwe);
 
-	if (!ecc_valid_point(&pwd->element_p))
-		goto invalid_point;
+	l_ecc_scalar_free(p_mask);
 
 	/* inv(p_mask * PWE) */
-	vli_sub(pwd->element_p.y, curve_p, pwd->element_p.y);
-
-	if (!ecc_valid_point(&pwd->element_p))
-		goto invalid_point;
-
-	/* change peer into to MSB first byte ordering before sending back */
-	ecc_native2be(pwd->element_p.x);
-	ecc_native2be(pwd->element_p.y);
-	ecc_native2be(pwd->scalar_p);
+	l_ecc_point_inverse(pwd->element_p);
 
 	/* send element_p and scalar_p */
 	pos = resp + 5; /* header */
 	*pos++ = EAP_PWD_EXCH_COMMIT;
-	memcpy(pos, pwd->element_p.x, ECC_BYTES);
-	pos += ECC_BYTES;
-	memcpy(pos, pwd->element_p.y, ECC_BYTES);
-	pos += ECC_BYTES;
-	memcpy(pos, pwd->scalar_p, ECC_BYTES);
-	pos += ECC_BYTES;
+	pos += l_ecc_point_get_data(pwd->element_p, pos, nbytes * 2);
+	pos += l_ecc_scalar_get_data(pwd->scalar_p, pos, nbytes);
 
 	eap_pwd_send_response(eap, resp, pos - resp);
 
@@ -466,18 +429,26 @@ error:
 	eap_method_error(eap);
 }
 
+
 static void eap_pwd_handle_confirm(struct eap_state *eap,
 					const uint8_t *pkt, size_t len)
 {
 	struct eap_pwd_handle *pwd = eap_get_data(eap);
-	struct ecc_point kp;
+	struct l_ecc_point *kp;
 	uint8_t resp[38];
 	uint8_t *pos;
-	uint64_t confirm_s[NUM_ECC_DIGITS];
-	uint8_t confirm_p[ECC_BYTES];
-	uint8_t expected_confirm_s[ECC_BYTES];
+	uint8_t confirm_s[32];
+	uint8_t confirm_p[32];
+	uint8_t expected_confirm_s[32];
 	uint8_t mk[32];
 	uint8_t msk_emsk[128], session_id[33];
+	/* buffers used for the final hash */
+	uint8_t kpx[L_ECC_SCALAR_MAX_BYTES];
+	uint8_t scalar_s[L_ECC_SCALAR_MAX_BYTES];
+	uint8_t scalar_p[L_ECC_SCALAR_MAX_BYTES];
+	uint8_t element_s[L_ECC_POINT_MAX_BYTES];
+	uint8_t element_p[L_ECC_POINT_MAX_BYTES];
+	ssize_t plen, clen;
 
 	if (len != 32) {
 		l_error("bad packet length");
@@ -491,75 +462,98 @@ static void eap_pwd_handle_confirm(struct eap_state *eap,
 
 	pwd->state = EAP_PWD_STATE_CONFIRM;
 
-	memcpy(confirm_s, pkt, ECC_BYTES);
+	memcpy(confirm_s, pkt, 32);
+
+	kp = l_ecc_point_new(pwd->curve);
 
 	/* compute KP = (p_rand * (Scalar_S * PWE + Element_S)) */
-	ecc_point_mult(&kp, &pwd->pwe, pwd->scalar_s, NULL,
-			vli_num_bits(pwd->scalar_s));
+	l_ecc_point_multiply(kp, pwd->scalar_s, pwd->pwe);
 
-	if (!ecc_valid_point(&kp))
+	l_ecc_point_add(kp, kp, pwd->element_s);
+
+	l_ecc_point_multiply(kp, pwd->p_rand, kp);
+
+	/*
+	 * We just need to store clen/plen once. Since all these buffers are
+	 * created with enough bytes in mind we know these wont fail. Also, all
+	 * scalar/point objects were created with the same curve, so it can be
+	 * safe to assume the return values will not change from what clen/plen
+	 * already are.
+	 */
+	clen = l_ecc_point_get_x(kp, kpx, sizeof(kpx));
+	if (clen < 0)
 		goto invalid_point;
 
-	ecc_point_add(&kp, &kp, &pwd->element_s);
-
-	if (!ecc_valid_point(&kp))
+	plen = l_ecc_point_get_data(pwd->element_s, element_s,
+					sizeof(element_s));
+	if (plen < 0)
 		goto invalid_point;
 
-	ecc_point_mult(&kp, &kp, pwd->p_rand, NULL, vli_num_bits(pwd->p_rand));
-
-	if (!ecc_valid_point(&kp))
+	if (l_ecc_point_get_data(pwd->element_p, element_p,
+					sizeof(element_p)) < 0)
 		goto invalid_point;
 
-	ecc_native2be(kp.x);
-	ecc_native2be(pwd->element_s.x);
-	ecc_native2be(pwd->element_s.y);
-	ecc_native2be(pwd->scalar_s);
+	if (l_ecc_scalar_get_data(pwd->scalar_s, scalar_s,
+					sizeof(scalar_s)) < 0)
+		goto invalid_point;
+
+	if (l_ecc_scalar_get_data(pwd->scalar_p, scalar_p,
+					sizeof(scalar_p)) < 0)
+		goto invalid_point;
+
+	l_ecc_point_free(kp);
 
 	/*
 	 * compute Confirm_P = H(kp | Element_P | Scalar_P |
 	 *                       Element_S | Scalar_S | Ciphersuite)
 	 */
-	H(8, confirm_p, kp.x, ECC_BYTES, pwd->element_p.x, ECC_BYTES,
-			pwd->element_p.y, ECC_BYTES, pwd->scalar_p,
-			ECC_BYTES, pwd->element_s.x, ECC_BYTES,
-			pwd->element_s.y, ECC_BYTES, pwd->scalar_s,
-			ECC_BYTES, &pwd->ciphersuite, 4);
+	hkdf_extract(L_CHECKSUM_SHA256, NULL, 0, 6, confirm_p, kpx, clen,
+				element_p, plen, scalar_p, clen, element_s,
+				plen, scalar_s, clen, &pwd->ciphersuite, 4);
 
-	H(8, expected_confirm_s, kp.x, ECC_BYTES, pwd->element_s.x,
-			ECC_BYTES, pwd->element_s.y, ECC_BYTES,
-			pwd->scalar_s, ECC_BYTES, pwd->element_p.x,
-			ECC_BYTES, pwd->element_p.y, ECC_BYTES,
-			pwd->scalar_p, ECC_BYTES, &pwd->ciphersuite, 4);
+	hkdf_extract(L_CHECKSUM_SHA256, NULL, 0, 6, expected_confirm_s, kpx,
+				clen, element_s, plen, scalar_s, clen,
+				element_p, plen, scalar_p, clen,
+				&pwd->ciphersuite, 4);
 
-	if (memcmp(confirm_s, expected_confirm_s, ECC_BYTES)) {
+	if (memcmp(confirm_s, expected_confirm_s, 32)) {
 		l_error("Confirm_S did not verify");
 		goto error;
 	}
 
 	pos = resp + 5; /* header */
 	*pos++ = EAP_PWD_EXCH_CONFIRM;
-	memcpy(pos, confirm_p, ECC_BYTES);
+	memcpy(pos, confirm_p, 32);
 	pos += 32;
 
 	/* derive MK = H(kp | Confirm_P | Confirm_S ) */
-	H(3, mk, kp.x, ECC_BYTES, confirm_p, ECC_BYTES,
-			confirm_s, ECC_BYTES);
+	hkdf_extract(L_CHECKSUM_SHA256, NULL, 0, 3, mk, kpx, clen, confirm_p,
+			32, confirm_s, 32);
 
 	eap_pwd_send_response(eap, resp, pos - resp);
 
 	eap_method_success(eap);
 
 	session_id[0] = 52;
-	H(3, session_id + 1, &pwd->ciphersuite, 4, pwd->scalar_p, ECC_BYTES,
-			pwd->scalar_s, ECC_BYTES);
+	hkdf_extract(L_CHECKSUM_SHA256, NULL, 0, 3, session_id + 1,
+			&pwd->ciphersuite, 4, scalar_p, clen, scalar_s, clen);
+
 	kdf(mk, 32, (const char *) session_id, 33, msk_emsk, 128);
 	eap_set_key_material(eap, msk_emsk, 64, msk_emsk + 64, 64, NULL, 0);
 
+	explicit_bzero(mk, sizeof(mk));
+	explicit_bzero(msk_emsk, sizeof(msk_emsk));
+	explicit_bzero(kpx, sizeof(kpx));
+
 	return;
 
 invalid_point:
+	l_ecc_point_free(kp);
+
 	l_error("invalid point during confirm exchange");
 error:
+	explicit_bzero(kpx, sizeof(kpx));
+
 	eap_method_error(eap);
 }
 
@@ -722,36 +716,56 @@ static int eap_pwd_check_settings(struct l_settings *settings,
 					const char *prefix,
 					struct l_queue **out_missing)
 {
-	const char *password;
-	L_AUTO_FREE_VAR(char *, identity);
 	const struct eap_secret_info *secret;
-	char setting[64], setting2[64];
+	char identity_key[72];
+	char password_key[72];
+	char password_key_old[72];
 
-	snprintf(setting, sizeof(setting), "%sIdentity", prefix);
-	identity = l_settings_get_string(settings, "Security", setting);
+	L_AUTO_FREE_VAR(char *, identity);
+	L_AUTO_FREE_VAR(char *, password) = NULL;
 
-	snprintf(setting2, sizeof(setting2), "%sPWD-Password", prefix);
-	password = l_settings_get_value(settings, "Security", setting2);
+	snprintf(identity_key, sizeof(identity_key), "%sIdentity", prefix);
+	snprintf(password_key, sizeof(password_key), "%sPassword", prefix);
+
+	identity = l_settings_get_string(settings, "Security", identity_key);
 
 	if (!identity) {
-		secret = l_queue_find(secrets, eap_secret_info_match, setting);
-		if (!secret) {
-			eap_append_secret(out_missing,
-					EAP_SECRET_REMOTE_USER_PASSWORD,
-					setting, setting2, NULL);
-		}
+		secret = l_queue_find(secrets, eap_secret_info_match,
+								identity_key);
+		if (secret)
+			return 0;
+
+		eap_append_secret(out_missing, EAP_SECRET_REMOTE_USER_PASSWORD,
+					identity_key, password_key, NULL,
+					EAP_CACHE_TEMPORARY);
 
 		return 0;
 	}
 
+	password = l_settings_get_string(settings, "Security", password_key);
+
 	if (!password) {
-		secret = l_queue_find(secrets, eap_secret_info_match, setting2);
-		if (!secret) {
-			eap_append_secret(out_missing,
-					EAP_SECRET_REMOTE_PASSWORD,
-					setting2, NULL, identity);
+		snprintf(password_key_old, sizeof(password_key_old),
+						"%sPWD-Password", prefix);
+		password = l_settings_get_string(settings, "Security",
+							password_key_old);
+		if (password) {
+			explicit_bzero(password, strlen(password));
+			l_warn("Setting '%s' is deprecated, use '%s' instead",
+					password_key_old, password_key);
+			return 0;
 		}
-	}
+
+		secret = l_queue_find(secrets, eap_secret_info_match,
+								password_key);
+		if (secret)
+			return 0;
+
+		eap_append_secret(out_missing, EAP_SECRET_REMOTE_PASSWORD,
+					password_key, NULL, identity,
+					EAP_CACHE_TEMPORARY);
+	} else
+		explicit_bzero(password, strlen(password));
 
 	return 0;
 }
@@ -761,26 +775,37 @@ static bool eap_pwd_load_settings(struct eap_state *eap,
 					const char *prefix)
 {
 	struct eap_pwd_handle *pwd;
-	char setting[64];
+	char setting_key[72];
 
 	pwd = l_new(struct eap_pwd_handle, 1);
 
 	pwd->state = EAP_PWD_STATE_INIT;
 
-	snprintf(setting, sizeof(setting), "%sIdentity", prefix);
-	pwd->identity = l_settings_get_string(settings, "Security", setting);
+	snprintf(setting_key, sizeof(setting_key), "%sIdentity", prefix);
+	pwd->identity = l_settings_get_string(settings, "Security",
+								setting_key);
 
 	if (!pwd->identity) {
-		l_error("EAP-Identity is missing");
+		l_error("'%s' setting is missing", setting_key);
 		goto error;
 	}
 
-	snprintf(setting, sizeof(setting), "%sPWD-Password", prefix);
-	pwd->password = l_settings_get_string(settings, "Security", setting);
+	snprintf(setting_key, sizeof(setting_key), "%sPassword", prefix);
+	pwd->password = l_settings_get_string(settings, "Security",
+								setting_key);
 
 	if (!pwd->password) {
-		l_error("EAP-PWD password is missing");
-		goto error;
+		snprintf(setting_key, sizeof(setting_key), "%sPWD-Password",
+									prefix);
+		pwd->password = l_settings_get_string(settings, "Security",
+								setting_key);
+
+		if (!pwd->password) {
+			snprintf(setting_key, sizeof(setting_key), "%sPassword",
+									prefix);
+			l_error("'%s' setting is missing", setting_key);
+			goto error;
+		}
 	}
 
 	eap_set_data(eap, pwd);
@@ -788,9 +813,14 @@ static bool eap_pwd_load_settings(struct eap_state *eap,
 	return true;
 
 error:
+	if (pwd->password) {
+		explicit_bzero(pwd->password, strlen(pwd->password));
+		l_free(pwd->password);
+	}
+
 	l_free(pwd->identity);
-	l_free(pwd->password);
 	l_free(pwd);
+
 	return false;
 }
 
