@@ -26,6 +26,7 @@
 #include <linux/types.h>
 
 struct handshake_state;
+enum crypto_cipher;
 
 /* 802.11-2016 Table 12-6 in section 12.7.2 */
 enum handshake_kde {
@@ -47,7 +48,8 @@ enum handshake_event {
 	HANDSHAKE_EVENT_SETTING_KEYS,
 	HANDSHAKE_EVENT_SETTING_KEYS_FAILED,
 	HANDSHAKE_EVENT_COMPLETE,
-	HANDSHAKE_EVENT_FAILED
+	HANDSHAKE_EVENT_FAILED,
+	HANDSHAKE_EVENT_REKEY_FAILED,
 };
 
 typedef void (*handshake_event_func_t)(struct handshake_state *hs,
@@ -77,8 +79,8 @@ struct handshake_state {
 	uint32_t ifindex;
 	uint8_t spa[6];
 	uint8_t aa[6];
-	uint8_t *ap_ie;
-	uint8_t *own_ie;
+	uint8_t *authenticator_ie;
+	uint8_t *supplicant_ie;
 	uint8_t *mde;
 	uint8_t *fte;
 	enum ie_rsn_cipher_suite pairwise_cipher;
@@ -86,9 +88,10 @@ struct handshake_state {
 	enum ie_rsn_cipher_suite group_management_cipher;
 	enum ie_rsn_akm_suite akm_suite;
 	uint8_t pmk[64];
+	size_t pmk_len;
 	uint8_t snonce[32];
 	uint8_t anonce[32];
-	uint8_t ptk[64];
+	uint8_t ptk[80];
 	uint8_t pmk_r0[32];
 	uint8_t pmk_r0_name[16];
 	uint8_t pmk_r1[32];
@@ -103,12 +106,17 @@ struct handshake_state {
 	bool have_anonce : 1;
 	bool have_pmkid : 1;
 	bool authenticator : 1;
+	bool wait_for_gtk : 1;
+	bool no_rekey : 1;
 	uint8_t ssid[32];
 	size_t ssid_len;
 	char *passphrase;
 	uint8_t r0khid[48];
 	size_t r0khid_len;
 	uint8_t r1khid[6];
+	uint8_t gtk[32];
+	uint8_t gtk_rsc[6];
+	unsigned int gtk_index;
 	void *user_data;
 
 	void (*free)(struct handshake_state *s);
@@ -127,14 +135,13 @@ void handshake_state_set_pmk(struct handshake_state *s, const uint8_t *pmk,
 				size_t pmk_len);
 void handshake_state_set_8021x_config(struct handshake_state *s,
 					struct l_settings *settings);
-struct l_settings *handshake_state_get_8021x_config(struct handshake_state *s);
-void handshake_state_set_ap_rsn(struct handshake_state *s,
+bool handshake_state_set_supplicant_rsn(struct handshake_state *s,
 					const uint8_t *rsn_ie);
-bool handshake_state_set_own_rsn(struct handshake_state *s,
+bool handshake_state_set_authenticator_rsn(struct handshake_state *s,
 					const uint8_t *rsn_ie);
-void handshake_state_set_ap_wpa(struct handshake_state *s,
+bool handshake_state_set_supplicant_wpa(struct handshake_state *s,
 					const uint8_t *wpa_ie);
-bool handshake_state_set_own_wpa(struct handshake_state *s,
+bool handshake_state_set_authenticator_wpa(struct handshake_state *s,
 					const uint8_t *wpa_ie);
 void handshake_state_set_ssid(struct handshake_state *s,
 					const uint8_t *ssid, size_t ssid_len);
@@ -151,6 +158,7 @@ void handshake_state_set_event_func(struct handshake_state *s,
 					void *user_data);
 void handshake_state_set_passphrase(struct handshake_state *s,
 					const char *passphrase);
+void handshake_state_set_no_rekey(struct handshake_state *s, bool no_rekey);
 
 void handshake_state_new_snonce(struct handshake_state *s);
 void handshake_state_new_anonce(struct handshake_state *s);
@@ -158,7 +166,9 @@ void handshake_state_set_anonce(struct handshake_state *s,
 				const uint8_t *anonce);
 void handshake_state_set_pmkid(struct handshake_state *s, const uint8_t *pmkid);
 bool handshake_state_derive_ptk(struct handshake_state *s);
-const struct crypto_ptk *handshake_state_get_ptk(struct handshake_state *s);
+size_t handshake_state_get_ptk_size(struct handshake_state *s);
+const uint8_t *handshake_state_get_kck(struct handshake_state *s);
+const uint8_t *handshake_state_get_kek(struct handshake_state *s);
 void handshake_state_install_ptk(struct handshake_state *s);
 
 void handshake_state_install_gtk(struct handshake_state *s,
@@ -179,6 +189,9 @@ bool handshake_state_get_pmkid(struct handshake_state *s, uint8_t *out_pmkid);
 bool handshake_decode_fte_key(struct handshake_state *s, const uint8_t *wrapped,
 				size_t key_len, uint8_t *key_out);
 
+void handshake_state_set_gtk(struct handshake_state *s, const uint8_t *key,
+				unsigned int key_index, const uint8_t *rsc);
+
 bool handshake_util_ap_ie_matches(const uint8_t *msg_ie,
 					const uint8_t *scan_ie, bool is_wpa);
 
@@ -188,6 +201,8 @@ const uint8_t *handshake_util_find_igtk_kde(const uint8_t *data,
 					size_t data_len, size_t *out_igtk_len);
 const uint8_t *handshake_util_find_pmkid_kde(const uint8_t *data,
 					size_t data_len);
+void handshake_util_build_gtk_kde(enum crypto_cipher cipher, const uint8_t *key,
+					unsigned int key_index, uint8_t *to);
 
 void handshake_event(struct handshake_state *hs, enum handshake_event event,
 		void *event_data);
