@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2015  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2015-2018  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -24,9 +24,11 @@
 #include <config.h>
 #endif
 
+#define _GNU_SOURCE
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
+#include <time.h>
 #include <sys/socket.h>
 #include <limits.h>
 #include <linux/if.h>
@@ -61,7 +63,6 @@ struct scan_periodic {
 	void *userdata;
 	bool rearm:1;
 	bool retry:1;
-	bool triggered:1;
 	bool needs_active_scan:1;
 	bool passive:1; /* Active or Passive scan? */
 	struct l_queue *cmds;
@@ -74,7 +75,6 @@ struct scan_request {
 	void *userdata;
 	scan_destroy_func_t destroy;
 	bool passive:1; /* Active or Passive scan? */
-	bool triggered:1;
 	struct l_queue *cmds;
 };
 
@@ -82,8 +82,10 @@ struct scan_context {
 	uint32_t ifindex;
 	enum scan_state state;
 	struct scan_periodic sp;
+	struct scan_request *current_sr;
 	struct l_queue *requests;
 	unsigned int start_cmd_id;
+	bool triggered:1;
 	struct wiphy *wiphy;
 };
 
@@ -117,18 +119,26 @@ static void scan_request_free(void *data)
 {
 	struct scan_request *sr = data;
 
+	if (sr->destroy)
+		sr->destroy(sr->userdata);
+
 	l_queue_destroy(sr->cmds, (l_queue_destroy_func_t) l_genl_msg_unref);
 
 	l_free(sr);
 }
 
-static void scan_request_trigger_failed(struct scan_request *sr, int err)
+static void scan_request_failed(struct scan_context *sc,
+				struct scan_request *sr, int err)
 {
+	sc->current_sr = NULL;
+	sc->triggered = false;
+	sc->start_cmd_id = 0;
+	l_queue_remove(sc->requests, sr);
+
 	if (sr->trigger)
 		sr->trigger(err, sr->userdata);
-
-	if (sr->destroy)
-		sr->destroy(sr->userdata);
+	else if (sr->callback)
+		sr->callback(err, NULL, sr->userdata);
 
 	scan_request_free(sr);
 }
@@ -208,24 +218,10 @@ bool scan_ifindex_remove(uint32_t ifindex)
 	return true;
 }
 
-static unsigned int scan_send_start(struct l_genl_msg **msg,
-			scan_func_t callback, void *user_data)
-{
-	unsigned int id = l_genl_family_send(nl80211, *msg, callback,
-						user_data, NULL);
-
-	if (id)
-		*msg = NULL;
-	else
-		l_error("Sending NL80211_CMD_TRIGGER_SCAN failed");
-
-	return id;
-}
-
-static void scan_triggered(struct l_genl_msg *msg, void *userdata)
+static void scan_request_triggered(struct l_genl_msg *msg, void *userdata)
 {
 	struct scan_context *sc = userdata;
-	struct scan_request *sr = l_queue_peek_head(sc->requests);
+	struct scan_request *sr = sc->current_sr;
 	int err;
 
 	l_debug("");
@@ -238,8 +234,7 @@ static void scan_triggered(struct l_genl_msg *msg, void *userdata)
 		if (err == -EBUSY)
 			return;
 
-		l_queue_pop_head(sc->requests);
-		scan_request_trigger_failed(sr, err);
+		scan_request_failed(sc, sr, err);
 
 		l_error("Received error during CMD_TRIGGER_SCAN: %s (%d)",
 			strerror(-err), -err);
@@ -252,17 +247,19 @@ static void scan_triggered(struct l_genl_msg *msg, void *userdata)
 	sc->state = sr->passive ? SCAN_STATE_PASSIVE : SCAN_STATE_ACTIVE;
 	l_debug("%s scan triggered for ifindex: %u",
 		sr->passive ? "Passive" : "Active", sc->ifindex);
-	sr->triggered = true;
+
+	sc->triggered = true;
+	l_genl_msg_unref(l_queue_pop_head(sr->cmds));
 
 	if (sr->trigger) {
 		sr->trigger(0, sr->userdata);
 
-        /*
-         * Reset callback for the consequent scan triggerings of the
-         * multi-segmented scans.
-         */
-        sr->trigger = NULL;
-    }
+		/*
+		 * Reset callback for the consequent scan triggerings of the
+		 * multi-segmented scans.
+		 */
+		sr->trigger = NULL;
+	}
 }
 
 struct scan_freq_append_data {
@@ -289,11 +286,17 @@ static void scan_build_attr_scan_frequencies(struct l_genl_msg *msg,
 	l_genl_msg_leave_nested(msg);
 }
 
-static void scan_freq_count(uint32_t freq, void *user_data)
+static bool scan_mac_address_randomization_is_disabled(void)
 {
-	int *count = user_data;
+	const struct l_settings *config = iwd_get_config();
+	bool disabled;
 
-	*count += 1;
+	if (!l_settings_get_bool(config, "Scan",
+					"disable_mac_address_randomization",
+					&disabled))
+		return false;
+
+	return disabled;
 }
 
 static struct l_genl_msg *scan_build_cmd(struct scan_context *sc,
@@ -301,16 +304,9 @@ static struct l_genl_msg *scan_build_cmd(struct scan_context *sc,
 					const struct scan_parameters *params)
 {
 	struct l_genl_msg *msg;
-	int n_channels = 0;
 	uint32_t flags = 0;
 
-	if (params->freqs)
-		scan_freq_set_foreach(params->freqs, scan_freq_count,
-					&n_channels);
-
-	msg = l_genl_msg_new_sized(NL80211_CMD_TRIGGER_SCAN,
-						64 + params->extra_ie_size +
-						4 * n_channels);
+	msg = l_genl_msg_new(NL80211_CMD_TRIGGER_SCAN);
 
 	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &sc->ifindex);
 
@@ -325,20 +321,14 @@ static struct l_genl_msg *scan_build_cmd(struct scan_context *sc,
 	if (params->flush && !ignore_flush_flag)
 		flags |= NL80211_SCAN_FLAG_FLUSH;
 
-	/*
-	 * TODO: Discovery of the hidden networks with randomization flag set
-	 * works with real hardware, but fails when used in simulated
-	 * environment with mac80211_hwsim. This needs to be investigated.
-	 *
-	 * if (params->randomize_mac_addr_hint &&
-	 *		wiphy_has_feature(sc->wiphy,
-	 *				  NL80211_FEATURE_SCAN_RANDOM_MAC_ADDR))
-	 *
-	 *	Randomizing 46 bits (locally administered 1 and multicast 0
-	 *	is assumed).
-	 *
-	 *	flags |= NL80211_SCAN_FLAG_RANDOM_ADDR;
-	*/
+	if (params->randomize_mac_addr_hint &&
+			wiphy_can_randomize_mac_addr(sc->wiphy) &&
+				!scan_mac_address_randomization_is_disabled())
+		/*
+		 * Randomizing 46 bits (locally administered 1 and multicast 0
+		 * is assumed).
+		 */
+		flags |= NL80211_SCAN_FLAG_RANDOM_ADDR;
 
 	if (flags)
 		l_genl_msg_append_attr(msg, NL80211_ATTR_SCAN_FLAGS, 4, &flags);
@@ -426,20 +416,27 @@ static void scan_cmds_add(struct l_queue *cmds, struct scan_context *sc,
 	l_queue_push_tail(cmds, cmd);
 }
 
-static int scan_request_send_next(struct scan_context *sc,
+static int scan_request_send_trigger(struct scan_context *sc,
 					struct scan_request *sr)
 {
-	struct l_genl_msg *cmd = l_queue_pop_head(sr->cmds);
+	struct l_genl_msg *cmd = l_queue_peek_head(sr->cmds);
 	if (!cmd)
 		return -ENOMSG;
 
-	sc->start_cmd_id = scan_send_start(&cmd, scan_triggered, sc);
+	sc->start_cmd_id = l_genl_family_send(nl80211, cmd,
+						scan_request_triggered, sc,
+									NULL);
 	if (sc->start_cmd_id) {
-		sr->triggered = false;
+		l_genl_msg_ref(cmd);
+
+		sc->triggered = false;
+		sc->current_sr = sr;
+
 		return 0;
 	}
 
-	l_genl_msg_unref(cmd);
+	l_error("Scan request: failed to trigger scan.");
+
 	return -EIO;
 }
 
@@ -472,13 +469,15 @@ static uint32_t scan_common(uint32_t ifindex, bool passive,
 	if (l_queue_length(sc->requests) > 0)
 		goto done;
 
-	if (sc->state != SCAN_STATE_NOT_RUNNING || sc->start_cmd_id)
+	if (sc->state != SCAN_STATE_NOT_RUNNING ||
+					sc->start_cmd_id || sc->triggered)
 		goto done;
 
-	if (!scan_request_send_next(sc, sr))
+	if (!scan_request_send_trigger(sc, sr))
 		goto done;
 
-	scan_request_trigger_failed(sr, -EIO);
+	sr->destroy = NULL;	/* Don't call destroy when returning error */
+	scan_request_free(sr);
 	return 0;
 done:
 	l_queue_push_tail(sc->requests, sr);
@@ -486,11 +485,11 @@ done:
 	return sr->id;
 }
 
-uint32_t scan_passive(uint32_t ifindex, scan_trigger_func_t trigger,
-			scan_notify_func_t notify, void *userdata,
-			scan_destroy_func_t destroy)
+uint32_t scan_passive(uint32_t ifindex, struct scan_freq_set *freqs,
+			scan_trigger_func_t trigger, scan_notify_func_t notify,
+			void *userdata, scan_destroy_func_t destroy)
 {
-	struct scan_parameters params = {};
+	struct scan_parameters params = { .freqs = freqs };
 
 	return scan_common(ifindex, true, &params, trigger, notify,
 							userdata, destroy);
@@ -530,24 +529,9 @@ bool scan_cancel(uint32_t ifindex, uint32_t id)
 	if (!sc)
 		return false;
 
-	sr = l_queue_peek_head(sc->requests);
-	if (!sr)
-		return false;
-
-	if (sr->id == id) {
-		/* If we already sent the trigger command, cancel the scan */
-		if (!sr->triggered && sc->start_cmd_id) {
-			l_genl_family_cancel(nl80211, sc->start_cmd_id);
-			sc->start_cmd_id = 0;
-
-			l_queue_pop_head(sc->requests);
-
-			start_next_scan_request(sc);
-
-			goto free;
-		}
-
-		/* If already triggered, just zero out the callback */
+	/* If already triggered, just zero out the callback */
+	if (sc->current_sr && sc->current_sr->id == id && sc->triggered) {
+		sr = sc->current_sr;
 		sr->callback = NULL;
 
 		if (sr->destroy) {
@@ -563,33 +547,18 @@ bool scan_cancel(uint32_t ifindex, uint32_t id)
 	if (!sr)
 		return false;
 
-free:
-	if (sr->destroy)
-		sr->destroy(sr->userdata);
+	/* If we already sent the trigger command, cancel the scan */
+	if (sr == sc->current_sr) {
+		l_genl_family_cancel(nl80211, sc->start_cmd_id);
+		sc->start_cmd_id = 0;
+		sc->current_sr = NULL;
+		start_next_scan_request(sc);
+	}
 
 	scan_request_free(sr);
-
 	return true;
 }
 
-void scan_sched_start(struct l_genl_family *nl80211, uint32_t ifindex,
-			uint32_t scan_interval,
-			scan_func_t callback, void *user_data)
-{
-	struct l_genl_msg *msg;
-
-	scan_interval *= 1000;	/* in kernel the interval is in msecs */
-
-	msg = l_genl_msg_new_sized(NL80211_CMD_START_SCHED_SCAN, 32);
-	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
-	l_genl_msg_append_attr(msg, NL80211_ATTR_SCHED_SCAN_INTERVAL,
-							4, &scan_interval);
-	l_genl_msg_append_attr(msg, NL80211_ATTR_SOCKET_OWNER, 0, NULL);
-
-	if (!l_genl_family_send(nl80211, msg, callback, user_data, NULL))
-		l_error("Starting scheduled scan failed");
-}
-
 static void scan_periodic_triggered(struct l_genl_msg *msg, void *user_data)
 {
 	struct scan_context *sc = user_data;
@@ -617,13 +586,13 @@ static void scan_periodic_triggered(struct l_genl_msg *msg, void *user_data)
 	l_debug("Periodic %s scan triggered for ifindex: %u", sc->sp.passive ?
 				"passive" : "active", sc->ifindex);
 
-	sc->sp.triggered = true;
+	sc->triggered = true;
 
 	if (sc->sp.trigger)
 		sc->sp.trigger(0, sc->sp.userdata);
 }
 
-static bool scan_periodic_send_start(struct scan_context *sc)
+static bool scan_periodic_send_trigger(struct scan_context *sc)
 {
 	struct l_genl_msg *cmd;
 	struct scan_parameters params = {};
@@ -643,7 +612,9 @@ static bool scan_periodic_send_start(struct scan_context *sc)
 	if (!cmd)
 		return false;
 
-	sc->start_cmd_id = scan_send_start(&cmd, scan_periodic_triggered, sc);
+	sc->start_cmd_id = l_genl_family_send(nl80211, cmd,
+						scan_periodic_triggered, sc,
+									NULL);
 	if (!sc->start_cmd_id) {
 		l_genl_msg_unref(cmd);
 		return false;
@@ -652,11 +623,26 @@ static bool scan_periodic_send_start(struct scan_context *sc)
 	return true;
 }
 
+static bool scan_periodic_is_disabled(void)
+{
+	const struct l_settings *config = iwd_get_config();
+	bool disabled;
+
+	if (!l_settings_get_bool(config, "Scan", "disable_periodic_scan",
+								&disabled))
+		return false;
+
+	return disabled;
+}
+
 void scan_periodic_start(uint32_t ifindex, scan_trigger_func_t trigger,
 				scan_notify_func_t func, void *userdata)
 {
 	struct scan_context *sc;
 
+	if (scan_periodic_is_disabled())
+		return;
+
 	sc = l_queue_find(scan_contexts, scan_context_match,
 				L_UINT_TO_PTR(ifindex));
 
@@ -744,21 +730,21 @@ static bool start_next_scan_request(struct scan_context *sc)
 {
 	struct scan_request *sr;
 
-	if (sc->state != SCAN_STATE_NOT_RUNNING || sc->start_cmd_id)
+	if (sc->state != SCAN_STATE_NOT_RUNNING ||
+					sc->start_cmd_id || sc->triggered)
 		return true;
 
 	while (!l_queue_isempty(sc->requests)) {
 		sr = l_queue_peek_head(sc->requests);
 
-		if (!scan_request_send_next(sc, sr))
+		if (!scan_request_send_trigger(sc, sr))
 			return true;
 
-		l_queue_pop_head(sc->requests);
-		scan_request_trigger_failed(sr, -EIO);
+		scan_request_failed(sc, sr, -EIO);
 	}
 
 	if (sc->sp.retry) {
-		if (scan_periodic_send_start(sc)) {
+		if (scan_periodic_send_trigger(sc)) {
 			sc->sp.retry = false;
 			return true;
 		}
@@ -788,11 +774,16 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 			have_ssid = true;
 			break;
 		case IE_TYPE_SUPPORTED_RATES:
+			if (iter.len > 8)
+				return false;
+
+			bss->has_sup_rates =  true;
+			memcpy(bss->supp_rates_ie, iter.data - 2, iter.len + 2);
+
+			break;
 		case IE_TYPE_EXTENDED_SUPPORTED_RATES:
-			if (ie_parse_supported_rates(&iter,
-						&bss->supported_rates) < 0)
-				l_warn("Unable to parse [Extended] "
-					"Supported Rates IE");
+			bss->ext_supp_rates_ie = l_memdup(iter.data - 2,
+								iter.len + 2);
 			break;
 		case IE_TYPE_RSN:
 			if (!bss->rsne)
@@ -802,7 +793,8 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 		case IE_TYPE_BSS_LOAD:
 			if (ie_parse_bss_load(&iter, NULL, &bss->utilization,
 						NULL) < 0)
-				l_warn("Unable to parse BSS Load IE");
+				l_warn("Unable to parse BSS Load IE for "
+					MAC, MAC_STR(bss->addr));
 			else
 				l_debug("Load: %u/255", bss->utilization);
 
@@ -838,6 +830,22 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 			bss->cc[2] = iter.data[2];
 			bss->cc_present = true;
 
+			break;
+		case IE_TYPE_HT_CAPABILITIES:
+			if (iter.len != 26)
+				return false;
+
+			bss->ht_capable = true;
+			memcpy(bss->ht_ie, iter.data - 2, iter.len + 2);
+
+			break;
+		case IE_TYPE_VHT_CAPABILITIES:
+			if (iter.len != 12)
+				return false;
+
+			bss->vht_capable = true;
+			memcpy(bss->vht_ie, iter.data - 2, iter.len + 2);
+
 			break;
 		}
 	}
@@ -972,17 +980,19 @@ static struct scan_bss *scan_parse_result(struct l_genl_msg *msg,
 	return bss;
 }
 
+/* User configurable options */
+static double RANK_5G_FACTOR;
+
 static void scan_bss_compute_rank(struct scan_bss *bss)
 {
 	static const double RANK_RSNE_FACTOR = 1.2;
 	static const double RANK_WPA_FACTOR = 1.0;
 	static const double RANK_OPEN_FACTOR = 0.5;
 	static const double RANK_NO_PRIVACY_FACTOR = 0.5;
-	static const double RANK_5G_FACTOR = 1.1;
 	static const double RANK_HIGH_UTILIZATION_FACTOR = 0.8;
 	static const double RANK_LOW_UTILIZATION_FACTOR = 1.2;
-	static const double RANK_MIN_SUPPORTED_RATE_FACTOR = 0.8;
-	static const double RANK_MAX_SUPPORTED_RATE_FACTOR = 1.1;
+	static const double RANK_MIN_SUPPORTED_RATE_FACTOR = 0.6;
+	static const double RANK_MAX_SUPPORTED_RATE_FACTOR = 1.3;
 	double rank;
 	uint32_t irank;
 
@@ -1019,17 +1029,27 @@ static void scan_bss_compute_rank(struct scan_bss *bss)
 	else if (bss->utilization <= 63)
 		rank *= RANK_LOW_UTILIZATION_FACTOR;
 
-	if (bss->supported_rates) {
-		uint8_t max = l_uintset_find_max(bss->supported_rates);
-		double factor = RANK_MAX_SUPPORTED_RATE_FACTOR -
+	if (bss->has_sup_rates || bss->ext_supp_rates_ie) {
+		uint64_t data_rate;
+
+		if (ie_parse_data_rates(bss->has_sup_rates ?
+					bss->supp_rates_ie : NULL,
+					bss->ext_supp_rates_ie,
+					bss->ht_capable ? bss->ht_ie : NULL,
+					bss->vht_capable ? bss->vht_ie : NULL,
+					bss->signal_strength / 100,
+					&data_rate) == 0) {
+			double factor = RANK_MAX_SUPPORTED_RATE_FACTOR -
 					RANK_MIN_SUPPORTED_RATE_FACTOR;
 
-		/*
-		 * Maximum rate is 54 Mbps, see DATA_RATE in 802.11-2012,
-		 * Section 6.5.5.2
-		 */
-		factor = factor * max / 108 + RANK_MIN_SUPPORTED_RATE_FACTOR;
-		rank *= factor;
+			/*
+			 * Maximum rate is 2340Mbps (VHT)
+			 */
+			factor = factor * data_rate / 2340000000U +
+						RANK_MIN_SUPPORTED_RATE_FACTOR;
+			rank *= factor;
+		} else
+			rank *= RANK_MIN_SUPPORTED_RATE_FACTOR;
 	}
 
 	irank = rank;
@@ -1042,14 +1062,14 @@ static void scan_bss_compute_rank(struct scan_bss *bss)
 
 void scan_bss_free(struct scan_bss *bss)
 {
-	l_uintset_free(bss->supported_rates);
+	l_free(bss->ext_supp_rates_ie);
 	l_free(bss->rsne);
 	l_free(bss->wpa);
 	l_free(bss->wsc);
 	l_free(bss);
 }
 
-int scan_bss_get_rsn_info(struct scan_bss *bss, struct ie_rsn_info *info)
+int scan_bss_get_rsn_info(const struct scan_bss *bss, struct ie_rsn_info *info)
 {
 	/*
 	 * If both an RSN and a WPA elements are present currently
@@ -1128,47 +1148,40 @@ static void discover_hidden_network_bsses(struct scan_context *sc,
 static void scan_finished(struct scan_context *sc, uint32_t wiphy,
 				int err, struct l_queue *bss_list)
 {
-	struct scan_request *sr;
+	struct scan_request *sr = sc->current_sr;
 	scan_notify_func_t callback = NULL;
 	void *userdata;
-	scan_destroy_func_t destroy = NULL;
 	bool new_owner = false;
 
-	sr = l_queue_peek_head(sc->requests);
-	if (sr && sr->triggered) {
+	if (sr) {
 		callback = sr->callback;
 		userdata = sr->userdata;
-		destroy = sr->destroy;
 
-		scan_request_free(sr);
+		sc->current_sr = NULL;
 		l_queue_pop_head(sc->requests);
 	} else if (sc->sp.interval) {
 		/*
 		 * If we'd called sc.sp->trigger, we must call back now
 		 * independent of whether the scan was succesful or was
 		 * aborted.  If the scan was successful though we call back
-		 * with the scan results even if didn't triggered this scan.
+		 * with the scan results even if we didn't trigger this scan.
 		 */
-		if (sc->sp.triggered || bss_list) {
+		if (sc->triggered || bss_list) {
 			callback = sc->sp.callback;
 			userdata = sc->sp.userdata;
-			destroy = NULL;
 		}
 
-		sc->sp.triggered = false;
-
 		if (bss_list)
 			discover_hidden_network_bsses(sc, bss_list);
 	}
 
 	if (callback)
-		new_owner = callback(wiphy, sc->ifindex, err,
-					bss_list, userdata);
+		new_owner = callback(err, bss_list, userdata);
 
-	if (destroy)
-		destroy(userdata);
+	if (sr)
+		scan_request_free(sr);
 
-	sc->state = SCAN_STATE_NOT_RUNNING;
+	sc->triggered = false;
 
 	if (!start_next_scan_request(sc) && sc->sp.rearm)
 		scan_periodic_rearm(sc);
@@ -1226,11 +1239,14 @@ static void scan_parse_new_scan_results(struct l_genl_msg *msg,
 
 static bool scan_send_next_cmd(struct scan_context *sc)
 {
-	struct scan_request *sr = l_queue_peek_head(sc->requests);
+	struct scan_request *sr = sc->current_sr;
 	int err;
 
-	if (sr && sr->triggered) {
-		err = scan_request_send_next(sc, sr);
+	if (!sc->triggered)
+		return false;
+
+	if (sr) {
+		err = scan_request_send_trigger(sc, sr);
 		if (!err)
 			return true;
 
@@ -1238,23 +1254,23 @@ static bool scan_send_next_cmd(struct scan_context *sc)
 		if (err < 0 && err == -ENOMSG)
 			return false;
 
-		sr = l_queue_pop_head(sc->requests);
-		scan_request_trigger_failed(sr, -EIO);
+		scan_request_failed(sc, sr, -EIO);
 
 		/*
 		 * The request is destroyed, return 'true' to stop further
 		 * processing.
 		 */
 		return true;
-	} else if (sc->sp.triggered) {
+	} else {
 		struct l_genl_msg *cmd = l_queue_pop_head(sc->sp.cmds);
 		if (!cmd)
 			return false;
 
-		sc->sp.triggered = false;
+		sc->triggered = false;
 
-		sc->start_cmd_id = scan_send_start(&cmd,
-						scan_periodic_triggered, sc);
+		sc->start_cmd_id = l_genl_family_send(nl80211, cmd,
+							scan_periodic_triggered,
+								sc, NULL);
 
 		if (!sc->start_cmd_id) {
 			l_genl_msg_unref(cmd);
@@ -1331,11 +1347,12 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 
 	switch (cmd) {
 	case NL80211_CMD_NEW_SCAN_RESULTS:
-	case NL80211_CMD_SCHED_SCAN_RESULTS:
 	{
 		struct l_genl_msg *scan_msg;
 		struct scan_results *results;
 
+		sc->state = SCAN_STATE_NOT_RUNNING;
+
 		if (scan_send_next_cmd(sc))
 			return;
 
@@ -1363,6 +1380,8 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 		break;
 
 	case NL80211_CMD_SCAN_ABORTED:
+		sc->state = SCAN_STATE_NOT_RUNNING;
+
 		scan_finished(sc, attr_wiphy, -ECANCELED, NULL);
 
 		break;
@@ -1638,29 +1657,69 @@ uint32_t scan_freq_set_get_bands(struct scan_freq_set *freqs)
 	return bands;
 }
 
-void scan_freq_set_foreach(struct scan_freq_set *freqs,
+static void scan_channels_5ghz_add(uint32_t channel, void *user_data)
+{
+	struct l_uintset *to = user_data;
+
+	l_uintset_put(to, channel);
+}
+
+void scan_freq_set_merge(struct scan_freq_set *to,
+					const struct scan_freq_set *from)
+{
+	to->channels_2ghz |= from->channels_2ghz;
+
+	l_uintset_foreach(from->channels_5ghz, scan_channels_5ghz_add,
+							to->channels_5ghz);
+}
+
+struct channels_5ghz_foreach_data {
+	scan_freq_set_func_t func;
+	void *user_data;
+};
+
+static void scan_channels_5ghz_frequency(uint32_t channel, void *user_data)
+{
+	const struct channels_5ghz_foreach_data *channels_5ghz_data = user_data;
+	uint32_t freq;
+
+	freq = scan_channel_to_freq(channel, SCAN_BAND_5_GHZ);
+
+	channels_5ghz_data->func(freq, channels_5ghz_data->user_data);
+}
+
+void scan_freq_set_foreach(const struct scan_freq_set *freqs,
 				scan_freq_set_func_t func, void *user_data)
 {
+	struct channels_5ghz_foreach_data data = { };
 	uint8_t channel;
 	uint32_t freq;
 
-	for (channel = 1; channel <= 14; channel++)
+	if (unlikely(!freqs || !func))
+		return;
+
+	data.func = func;
+	data.user_data = user_data;
+
+	l_uintset_foreach(freqs->channels_5ghz, scan_channels_5ghz_frequency,
+									&data);
+
+	if (!freqs->channels_2ghz)
+		return;
+
+	for (channel = 1; channel <= 14; channel++) {
 		if (freqs->channels_2ghz & (1 << (channel - 1))) {
 			freq = scan_channel_to_freq(channel, SCAN_BAND_2_4_GHZ);
 
 			func(freq, user_data);
 		}
-
-	for (channel = 1; channel <= 200; channel++)
-		if (l_uintset_contains(freqs->channels_5ghz, channel)) {
-			freq = scan_channel_to_freq(channel, SCAN_BAND_5_GHZ);
-
-			func(freq, user_data);
-		}
+	}
 }
 
 bool scan_init(struct l_genl_family *in)
 {
+	const struct l_settings *config = iwd_get_config();
+
 	nl80211 = in;
 	scan_id = l_genl_family_register(nl80211, "scan", scan_notify,
 						NULL, NULL);
@@ -1672,6 +1731,10 @@ bool scan_init(struct l_genl_family *in)
 
 	scan_contexts = l_queue_new();
 
+	if (!l_settings_get_double(config, "Rank", "rank_5g_factor",
+					&RANK_5G_FACTOR))
+		RANK_5G_FACTOR = 1.0;
+
 	return true;
 }
 
